from typing import Callable, List

from discord import Embed


class Page(Embed):
    """Represents a single page inside a menu."""

    __slots__ = (*Embed.__slots__, '_index', '_buttons_list', '_on_next_event', '_on_fail_event', '_on_cancel_event', '_on_timeout_event')

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __repr__(self):
        return f"Page(title={self.title} " \
               f"{''.join([f'{k}={v}' for k, v in {j: getattr(self, j) for j in self.__slots__}])})"

    def __str__(self):
        return f'Page {self.index} {self.title}'

    @property
    def index(self) -> int:
        return getattr(self, '_index', 0)

    @index.setter
    def index(self, i: int):
        self._index = i

    @property
    def buttons_list(self) -> List:
        return getattr(self, '_buttons_list', [])

    def buttons(self, buttons: List) -> 'Page':
        """Generates reaction buttons when the page is displayed. Returns itself for fluent-style chaining."""
        self._buttons_list = buttons

        return self

    @property
    def on_next_event(self) -> Callable:
        return getattr(self, '_on_next_event', None)

    def on_next(self, func: Callable) -> 'Page':
        """Sets the function that will be called when the `next` event runs. Returns itself for fluent-style chaining."""
        self._on_next_event = func

        return self

    @property
    def on_fail_event(self) -> Callable:
        return getattr(self, '_on_fail_event', None)

    def on_fail(self, func: Callable) -> 'Page':
        """Sets the function that will be called when the `fail` event runs. Returns itself for fluent-style chaining."""
        self._on_fail_event = func

        return self

    @property
    def on_cancel_event(self) -> Callable:
        return getattr(self, '_on_cancel_event', None)

    def on_cancel(self, func: Callable) -> 'Page':
        """Sets the function that will be called when the `cancel` event runs. Returns itself for fluent-style chaining."""
        self._on_cancel_event = func

        return self

    @property
    def on_timeout_event(self) -> Callable:
        return getattr(self, '_on_timeout_event', None)

    def on_timeout(self, func: Callable) -> 'Page':
        """Sets the function that will be called when the `timeout` event runs. Returns itself for fluent-style chaining."""
        self._on_timeout_event = func

        return self

    def as_safe_embed(self) -> 'Page':
        """Returns a page stripped of Callables and Page-specific properties so we can send it as a standard Embed."""
        safe_embed = self.to_dict()
        return Embed.from_dict(safe_embed)



import asyncio
import logging
from typing import Dict, List, Optional, TypeVar

from discord import Embed, Emoji, Message, PartialEmoji, RawReactionActionEvent, Reaction
from discord.abc import GuildChannel
from discord.ext.commands import Context

from dpymenus import ButtonMenu, Page
from dpymenus.base_menu import sessions
from dpymenus.constants import GENERIC_BUTTONS
from dpymenus.exceptions import ButtonsError, PagesError, SessionError

Button = TypeVar('Button', Emoji, PartialEmoji, str)
PageType = TypeVar('PageType', Embed, Page, Dict)


class PaginatedMenu(ButtonMenu):
    """
    Represents an paginated, button-based response menu.

    :param ctx: A reference to the command context.
    """

    def __init__(self, ctx: Context):
        super().__init__(ctx)

    def __repr__(self):
        return f'PaginatedMenu(pages={[p.__str__() for p in self.pages]}, page={self.page}, timeout={self.timeout}, ' \
               f'skip_buttons={self.skip_buttons} page_numbers={self.page_numbers}, cancel_page={self.cancel_page}, ' \
               f'timeout_page={self.timeout_page})'

    @property
    def cancel_page(self) -> Optional[Embed]:
        return getattr(self, '_cancel_page', None)

    def set_cancel_page(self, embed: Embed) -> 'PaginatedMenu':
        """Sets the function that will be called when the `cancel` event runs. Returns itself for fluent-style chaining."""
        self._cancel_page = embed

        return self

    @property
    def timeout_page(self) -> Optional[Embed]:
        return getattr(self, '_timeout_page', None)

    def set_timeout_page(self, embed: Embed) -> 'PaginatedMenu':
        """Sets the function that will be called when the `cancel` event runs. Returns itself for fluent-style chaining."""
        self._timeout_page = embed

        return self

    @property
    def page_numbers(self) -> bool:
        return getattr(self, '_page_numbers', False)

    def show_page_numbers(self) -> 'PaginatedMenu':
        """Adds page numbers to each embeds by overwriting the footer. Returns itself for fluent-style chaining."""
        self._page_numbers = True

        # check if pages were set prior to calling this method; if so, we reset footers
        if self.pages:
            for i, page in enumerate(self.pages):
                page.set_footer(text=f'{i + 1}/{len(self.pages)}')

        return self

    @property
    def skip_buttons(self) -> bool:
        return getattr(self, '_skip_buttons', False)

    def show_skip_buttons(self) -> 'PaginatedMenu':
        """Adds two extra buttons for jumping to the first and last page. Returns itself for fluent-style chaining."""
        self._skip_buttons = True

        return self

    @property
    def cancel_button(self) -> bool:
        return getattr(self, '_cancel_button', True)

    def hide_cancel_button(self) -> 'PaginatedMenu':
        """Sets whether to show the cancel button or not. Returns itself for fluent-style chaining."""
        self._cancel_button = False

        return self

    @property
    def buttons_list(self) -> List:
        return getattr(self, '_buttons_list', [])

    def buttons(self, buttons: List[Button]) -> 'PaginatedMenu':
        """Replaces the default buttons. You must include 3 or 5 emoji/strings in the order they would be displayed.
        0 and 5 are only shown if `enable_skip_buttons` is set, otherwise 2, 3, and 4 will be shown. You can pass in
        `None` or an empty string for 0 and 5 if you do not intend on using them. If you only pass in 3 values, they
         will be filled in as the defaults for you. If you enable the skip buttons without having values set, it will
         use those defaults."""
        self._buttons_list = buttons

        if len(buttons) == 3:
            self.buttons_list.insert(0, GENERIC_BUTTONS[0])
            self.buttons_list.insert(4, GENERIC_BUTTONS[4])

        return self

    @property
    def prevent_multisessions(self) -> bool:
        return getattr(self, '_prevent_multisessions', True)

    def allow_multisession(self):
        """Sets whether or not a user can open a new menu when one is already open. This will close the old menu.
        Returns itself for fluent-style chaining."""
        self._prevent_multisessions = False

        return self

    async def open(self):
        """The entry point to a new PaginatedMenu instance; starts the main menu loop.
        Manages gathering user input, basic validation, sending messages, and cancellation requests."""
        if not self.prevent_multisessions:
            if (self.ctx.author.id, self.ctx.channel.id) in sessions.keys():
                session: 'PaginatedMenu' = sessions.get((self.ctx.author.id, self.ctx.channel.id))
                await session._cleanup_reactions()
                await session.close_session()

        try:
            # we have to set these here so we can validate cleanly
            if len(self.buttons_list) == 0:
                self.buttons(GENERIC_BUTTONS)

            self._validate_buttons()
            await super()._open()

        except (ButtonsError, PagesError) as exc:
            logging.error(exc.message)

        except SessionError as exc:
            logging.info(exc.message)

        else:
            await self._add_buttons()

            # refresh our message content with the reactions added
            self.output = await self.destination.fetch_message(self.output.id)

            while self.active:
                tasks = [asyncio.create_task(self._get_reaction_add()), asyncio.create_task(self._get_reaction_remove())]

                if not self.prevent_multisessions:
                    tasks.append(asyncio.create_task(self._shortcircuit()))

                done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED, timeout=self.timeout)

                # if all tasks are still pending, we force a timeout by manually calling cleanup methods
                if len(pending) == len(tasks):
                    await self._execute_timeout()

                else:
                    for future in done:
                        result = future.result()
                        if result:
                            self.input = result
                            break

                        else:
                            return

                    if isinstance(self.output.channel, GuildChannel):
                        await self.output.remove_reaction(self.input, self.ctx.author)

                    await self._handle_transition()

                for task in pending:
                    task.cancel()

            await self._cleanup_reactions()

    async def send_message(self, embed: Embed) -> Message:
        """
        Edits the menu output message. We override the :class:`~dpymenus.BaseMenu` implementation because
        we always want to edit, even in a DM channel type.

        :param embed: A Discord :py:class:`~discord.Embed` object.
        """
        return await self.output.edit(embed=embed)

    def add_pages(self, pages: List[PageType]) -> 'PaginatedMenu':
        """Helper method to convert embeds into Pages and add them to a menu."""
        self._validate_pages(pages)

        for i, page in enumerate(pages):
            if isinstance(page, dict):
                page = Page.from_dict(page)

            # explicit type check here because Pages are instances of Embeds
            elif type(page) == Embed:
                page = Page.from_dict(page.to_dict())

            if self.page_numbers:
                page.set_footer(text=f'{i + 1}/{len(pages)}')

            page.index = i
            self.pages.append(page)

        self.page = self.pages[0]

        return self

    # Internal Methods
    async def _execute_cancel(self):
        """Sends a cancellation message. Deletes the menu message if no page was set."""
        cancel_page = getattr(self, 'cancel_page', None)

        if cancel_page:
            await self.output.edit(embed=cancel_page)

        else:
            await self._cleanup_output()

        await self.close_session()

    async def _execute_timeout(self):
        """Sends a timeout message. Deletes the menu message if no page was set."""
        try:
            await self.close_session()

        except KeyError:
            return

        timeout_page = getattr(self, 'timeout_page')

        if timeout_page:
            await self.output.edit(embed=timeout_page)

        else:
            await self._cleanup_output()

        self.active = False

    async def _shortcircuit(self):
        """Runs a background loop to poll the menus `active` state. Returns when False. Allows for short-circuiting the main
        loop when it is waiting for user reaction events from discord.py."""
        while self.active:
            await asyncio.sleep(1)

        else:
            return

    async def _get_reaction_add(self) -> Button:
        """Collects a user reaction and places it into the input attribute. Returns a :py:class:`discord.Emoji` or string."""
        reaction_event = await self.ctx.bot.wait_for('raw_reaction_add', check=self._check_reaction if self.buttons_list != GENERIC_BUTTONS else self._check_reaction_defaults)

        return reaction_event.emoji

    async def _get_reaction_remove(self) -> Button:
        """Collects a user reaction and places it into the input attribute. Returns a :py:class:`discord.Emoji` or string."""
        reaction_event = await self.ctx.bot.wait_for('raw_reaction_remove', check=self._check_reaction if self.buttons_list != GENERIC_BUTTONS else self._check_reaction_defaults)

        return reaction_event.emoji

    def _check_reaction_defaults(self, event: RawReactionActionEvent) -> bool:
        """Returns true if the author is the person who reacted and the message ID's match. Checks the generic buttons."""
        return (event.user_id == self.ctx.author.id
                and event.message_id == self.output.id
                and any(event.emoji.name == btn for btn in self.buttons_list))

    def _validate_buttons(self):
        if self.buttons_list != GENERIC_BUTTONS:
            if len(self.buttons_list) != 3 and len(self.buttons_list) != 5:
                raise ButtonsError(f'Buttons length mismatch. Expected 3 or 5, found {len(self.buttons_list)}')

            self._check_buttons(self.buttons_list)

    async def _add_buttons(self):
        """Adds reactions to the message object based on what was passed into the page buttons."""
        if not self.cancel_button:
            self.buttons_list[2] = None

        if self.skip_buttons:
            for button in self.buttons_list:
                if button is None:
                    continue

                await self.output.add_reaction(button)

        else:
            for button in self.buttons_list[1:4]:
                if button is None:
                    continue

                await self.output.add_reaction(button)

    async def _handle_transition(self):
        """Dictionary mapping of reactions to methods to be called when handling user input on a button."""
        transitions = [self.to_first, self.previous, self.close, self.next, self.to_last]

        if not self.cancel_button:
            transitions.remove(self.close)

        if not self.skip_buttons:
            transitions.remove(self.to_first)
            transitions.remove(self.to_last)

        # cursed code, not sure how else to cover all cases though; watch for performance issues
        transition_map = {(button.emoji.name if isinstance(button.emoji, Emoji) else button.emoji)
                          if isinstance(button, Reaction) else button: transition
                          for button, transition in zip(self.output.reactions, transitions)}

        await transition_map[self.input.name]()



import asyncio
import logging
from typing import Dict, Set
from warnings import warn

from discord import RawReactionActionEvent, User
from discord.ext.commands import Context

from dpymenus import ButtonMenu
from dpymenus.exceptions import ButtonsError, EventError, PagesError, SessionError


class Poll(ButtonMenu):
    """Represents a Poll menu.

    :param ctx: A reference to the command context.
    """

    def __init__(self, ctx: Context):
        super().__init__(ctx)
        self.voted: Set[User] = set()

    def __repr__(self):
        return f'Poll(pages={[p.__str__() for p in self.pages]}, page={self.page.index}, timeout={self.timeout}, data={self.data})'

    async def open(self):
        """The entry point to a new Poll instance; starts the main menu loop.
        Manages gathering user input, basic validation, sending messages, and cancellation requests."""
        try:
            await super()._open()
        except SessionError as exc:
            logging.info(exc.message)
        else:
            await self._set_data()
            await self._add_buttons()

            pending = set()
            while self.active:
                try:
                    _, pending = await asyncio.wait(
                            [
                                asyncio.create_task(self._get_vote_add()),
                                asyncio.create_task(self._get_vote_remove()),
                                asyncio.create_task(self._poll_timer())
                             ],
                            return_when=asyncio.FIRST_COMPLETED)

                finally:
                    for task in pending:
                        task.cancel()

                    await self._finish_poll()

    # Utility Methods
    async def results(self) -> Dict[str, int]:
        """Utility method to get a dictionary of poll results."""
        return {choice: len(voters) for choice, voters in self.data.items()}

    async def add_results_fields(self):
        """Utility method to add new fields to your next page automatically."""
        for choice, voters in self.data.items():
            next_page = self.pages[self.page.index + 1]
            next_page.add_field(name=choice, value=str(len(voters)))

    async def generate_results_page(self):
        """Utility method to build your entire results page automatically."""
        next_page = self.pages[self.page.index + 1]

        await self.add_results_fields()

        highest_value = max(self.data.values())
        winning_key = {choice for choice, voters in self.data.items() if voters == highest_value}

        if len(highest_value) == 0:
            next_page.description = ' '.join([next_page.description, f"It's a draw!"])

        else:
            next_page.description = ' '.join([next_page.description, f'{str(next(iter(winning_key)))} wins!'])

    # Internal Methods
    async def _get_vote_add(self):
        """Watches for a user adding a reaction on the Poll. Adds them to the relevant state_field values."""
        while True:
            try:
                reaction_event = await self.ctx.bot.wait_for('raw_reaction_add', timeout=self.timeout,
                                                             check=self._check_reaction)

            except asyncio.TimeoutError:
                return

            else:
                if reaction_event.emoji.name in self.page.buttons_list:
                    self.data[reaction_event.emoji.name].add(reaction_event.user_id)

    async def _get_vote_remove(self):
        """Watches for a user removing a reaction on the Poll. Removes them from the relevant state_field values."""
        while True:
            try:
                reaction_event = await self.ctx.bot.wait_for('raw_reaction_remove', timeout=self.timeout,
                                                             check=self._check_reaction)

            except asyncio.TimeoutError:
                return

            else:
                if reaction_event.emoji.name in self.page.buttons_list:
                    self.data[reaction_event.emoji.name].remove(reaction_event.user_id)

    def _check_reaction(self, event: RawReactionActionEvent) -> bool:
        """Returns true only if the reaction event member is not a bot (ie. excludes self from counts)."""
        return event.member is not None and event.member.bot is False

    async def _poll_timer(self):
        """Handles poll duration."""
        await asyncio.sleep(self.timeout)

    async def _finish_poll(self):
        """Removes multi-votes and calls the Page on_next function when finished."""
        cheaters = await self._get_cheaters()
        for voters in self.data.values():
            voters -= cheaters

        await self.output.clear_reactions()
        await self.page.on_next_event(self)

    async def _get_cheaters(self) -> Set[int]:
        """Returns a set of user ID's that appear in more than one state_field value."""
        seen = set()
        repeated = set()
        for voters in self.data.values():
            for voter in voters:
                repeated.add(voter) if voter in seen else seen.add(voter)

        return repeated

    async def _set_data(self):
        """Internally sets data field keys and values based on the current Page button properties."""
        self._validate_buttons()

        self.set_data({})
        for button in self.page.buttons_list:
            self.data.update({button: set()})

    def _validate_buttons(self):
        """Checks that Poll objects always have more than two buttons."""
        if len(self.page.buttons_list) < 2:
            raise ButtonsError(f'A Poll primary page must have at least two buttons. Expected at least 2, found {len(self.page.buttons_list)}.')

    def _validate_pages(self):
        """Checks that the Menu contains at least one Page."""
        if len(self.pages) != 2:
            raise PagesError(f'A Poll can only have two pages. Expected 2, found {len(self.pages)}.')

        if self.page.on_cancel_event or self.page.on_fail_event or self.page.on_timeout_event:
            raise EventError('A Poll can not capture a `cancel`, `fail`, or `timeout` event.')

        if len(self.page.buttons_list) > 5:
            warn('Adding more than 5 buttons to a page at once may result in discord.py throttling the bot client.')

    @staticmethod
    async def get_voters(users: Set[User]) -> Set[User]:
        """Returns a set of user ID's."""
        return {user.id for user in users}



import asyncio
import logging
from typing import Dict, Optional, TypeVar

import emoji
from discord import Emoji, PartialEmoji, RawReactionActionEvent, Reaction
from discord.abc import GuildChannel
from discord.ext.commands import Context

from dpymenus import BaseMenu
from dpymenus.exceptions import ButtonsError, EventError, SessionError

Button = TypeVar('Button', Emoji, PartialEmoji, str)


class ButtonMenu(BaseMenu):
    """
    Represents a button-based response menu.

    :param ctx: A reference to the command context.
    """

    def __init__(self, ctx: Context):
        super().__init__(ctx)

    def __repr__(self):
        return f'ButtonMenu(pages={[p.__str__() for p in self.pages]}, page={self.page}, timeout={self.timeout}, data={self.data})'

    @property
    def data(self) -> Dict:
        return getattr(self, '_data', {})

    def set_data(self, data: Dict) -> 'ButtonMenu':
        """Sets a dictionary up for persistent state data. Returns itself for fluent-style chaining."""
        self._data = data

        return self

    def button_pressed(self, button: Button) -> bool:
        """Helper method which checks if the button the user pressed is the button passed in."""
        return button == self.input

    async def open(self):
        """The entry point to a new ButtonMenu instance; starts the main menu loop.
        Manages gathering user input, basic validation, sending messages, and cancellation requests."""
        try:
            self._validate_buttons()
            await super()._open()

        except (ButtonsError, EventError) as exc:
            logging.error(exc.message)

        except SessionError as exc:
            logging.info(exc.message)

        else:
            await self._add_buttons()
            _first_iter = True

            while self.active:
                if not _first_iter and self.last_visited_page() != self.page.index:
                    await self._add_buttons()

                elif not _first_iter and self.last_visited_page() == self.page.index:
                    if isinstance(self.output.channel, GuildChannel):
                        await self.output.remove_reaction(self.input, self.ctx.author)

                # refresh our message content with the reactions added
                self.output = await self.destination.fetch_message(self.output.id)

                self.input = await self._get_reaction_add()

                await self.page.on_next_event(self)

                if self.last_visited_page() != self.page.index:
                    await self._cleanup_reactions()

                _first_iter = False

    # Internal Methods
    async def _add_buttons(self):
        """Adds reactions to the message object based on what was passed into the page buttons."""
        for button in self.page.buttons_list:
            await self.output.add_reaction(button)

    async def _get_reaction_add(self) -> Optional[Button]:
        """Collects a user reaction and places it into the input attribute. Returns a :py:class:`discord.Emoji` or string."""
        try:
            reaction_event = await self.ctx.bot.wait_for('raw_reaction_add', check=self._check_reaction)

        except asyncio.TimeoutError:
            await self._execute_timeout()

        else:
            for btn in self.page.buttons_list:
                if isinstance(btn, Emoji):
                    if btn == reaction_event.emoji:
                        return btn

                elif isinstance(btn, str):
                    # split the str and test if the value between ':' is the same as the PartialEmoji name
                    _test = btn.split(':')
                    if len(_test) > 1:
                        if _test[1] == reaction_event.emoji.name:
                            return btn

                    else:
                        if btn == reaction_event.emoji.name:
                            return btn

                else:
                    return reaction_event.emoji

    async def _cleanup_reactions(self):
        """Removes all reactions from the output message object."""
        if self.output and isinstance(self.output.channel, GuildChannel):
            await self.output.clear_reactions()

    def _check_reaction(self, event: RawReactionActionEvent) -> bool:
        """Returns true if the author is the person who reacted and the message ID's match. Checks the pages buttons."""
        # cursed code, not sure how else to cover all cases though; watch for performance issues
        return (event.user_id == self.ctx.author.id
                and event.message_id == self.output.id
                and any(event.emoji.name == btn
                        for btn in [(reaction.emoji.name if isinstance(reaction.emoji, Emoji) else reaction.emoji)
                                    if isinstance(reaction, Reaction) else reaction
                                    for reaction in self.output.reactions]))

    def _validate_buttons(self):
        """Ensures that a menu was passed the appropriate amount of buttons."""
        _cb_count = 0
        for page in self.pages:
            if not page.buttons_list:
                break

            if page.on_next_event:
                _cb_count += 1

            if len(page.buttons_list) < 1:
                raise ButtonsError('Any page with an `on_next` event capture must have at least one button.\n'
                                   f'{page} {page.title} only has {len(page.buttons_list)} buttons.')

            if len(page.buttons_list) > 5:
                logging.warning('Adding more than 5 buttons to a page at once may result in discord.py throttling the bot client.')

            self._check_buttons(page.buttons_list)

        if self.page.on_fail_event:
            raise EventError('A ButtonMenu can not capture an `on_fail` event.')

        if _cb_count < len(self.pages) - 1:
            raise EventError(f'ButtonMenu missing `on_next` captures. Expected {len(self.pages) - 1}, found {_cb_count}.')

    def _check_buttons(self, buttons_list):
        for button in buttons_list:
            if isinstance(button, (Emoji, PartialEmoji)):
                continue

            if isinstance(button, str):
                # split the str and test if the value between ':' is in the bot list
                _test = button.split(':')
                if len(_test) > 1:
                    if _test[1] in [e.name for e in self.ctx.bot.emojis]:
                        continue

                # check by key; faster than iterating over the list w/ for loop
                _test = emoji.UNICODE_EMOJI_ALIAS.get(button, None)
                if _test:
                    continue

            raise ButtonsError(f'Invalid Emoji or unicode string: {button}')



import abc
import asyncio
import logging
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union

from discord import Embed, Message, Reaction, TextChannel, User
from discord.abc import GuildChannel
from discord.ext.commands import Context

from dpymenus.exceptions import PagesError, SessionError
from dpymenus.page import Page

PageType = TypeVar('PageType', Embed, Page, Dict)

sessions: Dict[Tuple[int, int], Any]
sessions = dict()

logger = logging.getLogger('dpymenus')
logger.addHandler(logging.NullHandler())


class BaseMenu(abc.ABC):
    """Represents the base menu from which TextMenu, ButtonMenu, and Poll inherit from.

    Attributes
        :ctx: A reference to the command Context.
        :pages: A list containing references to Page objects.
        :page: Current Page object.
        :active: Whether or not the menu is active or not.
        :input: A reference to the captured user input message object.
        :output: A reference to the menus output message.
        :history: An ordered history of pages visited by the user.
    """

    def __init__(self, ctx: Context):
        self.ctx: Context = ctx
        self.pages: List[Page] = []
        self.page: Optional[Page] = None
        self.active: bool = True
        self.input: Optional[Union[Message, Reaction]] = None
        self.output: Optional[Message] = None
        self.history: List[int] = []

    @abc.abstractmethod
    async def open(self):
        """The entry point to a new Menu instance; starts the main menu loop.
        Manages gathering user input, basic validation, sending messages, and cancellation requests."""
        pass

    @property
    def timeout(self) -> int:
        return getattr(self, '_timeout', 300)

    def set_timeout(self, timeout: int) -> 'BaseMenu':
        """Sets the timeout duration (in seconds) for the menu. Returns itself for fluent-style chaining."""
        self._timeout = timeout

        return self

    @property
    def destination(self) -> Union[Context, User, TextChannel]:
        return getattr(self, '_destination', self.ctx)

    def set_destination(self, dest: Union[User, TextChannel]) -> 'BaseMenu':
        """Sets the message destination for the menu. Returns itself for fluent-style chaining."""
        self._destination = dest

        return self

    @property
    def command_message(self) -> bool:
        return getattr(self, '_command_message', False)

    def show_command_message(self) -> 'BaseMenu':
        """Persists user command invocation messages in the chat instead of deleting them after execution."""
        self._command_message = True

        return self

    @property
    def persist(self) -> bool:
        return getattr(self, '_persist', False)

    def persist_on_close(self) -> 'BaseMenu':
        """Prevents message cleanup from running when a menu closes."""
        self._persist = True

        return self

    async def close(self):
        """Helper method to close the menu out properly. Used to manually call a cancel event."""
        await self._execute_cancel()

    async def next(self):
        """Sets a specific :class:`~dpymenus.Page` to go to and calls the :func:`~send_message()` method to display the embed."""
        if self.page.index + 1 > len(self.pages) - 1:
            return

        self.page = self.pages[self.page.index + 1]

        await self._post_next()

    async def previous(self):
        """Helper method for quickly accessing the previous page."""
        if self.page.index - 1 < 0:
            return

        self.page = self.pages[self.page.index - 1]

        await self._post_next()

    async def to_first(self):
        """Helper method to jump to the first page."""
        self.page = self.pages[0]

        await self._post_next()

    async def to_last(self):
        """Helper method to jump to the last page."""
        self.page = self.pages[-1:][0]

        await self._post_next()

    async def go_to(self, page: Optional[Union[str, int]] = None):
        """Sets a specific :class:`~dpymenus.Page` to go to and calls the :func:`~send_message()` method to display the embed.

        :param page: The name of the `on_next` function for a particular page or its page number. If this is not set, the next
        page in the list will be called.
        """
        if isinstance(page, int):
            self.page = self.pages[page]

        elif isinstance(page, str):
            # get a page index from its on_next callback function name and assign it
            for p in self.pages:
                if p.on_next_event.__name__ == page:
                    self.page = p
                    break

        await self._post_next()

    def last_visited_page(self) -> int:
        """Returns the last visited pages index."""
        return self.history[-2] if len(self.history) > 1 else 0

    def add_pages(self, pages: List[PageType]) -> 'BaseMenu':
        """Adds a list of pages to a menu, setting their index based on the position in the list."""
        self._validate_pages(pages)

        for i, page in enumerate(pages):
            if type(page) == Embed:
                page = Page.from_dict(page.to_dict())

            page.index = i
            self.pages.append(page)

        self.page = self.pages[0]

        return self

    async def send_message(self, page: PageType) -> Message:
        """
        Edits a message if the channel is in a Guild, otherwise sends it to the current channel.

        :param page: A Discord :py:class:`~discord.Embed` or :class:`~dpymenus.Page` object.
        """
        safe_embed = page.as_safe_embed() if type(page) == Page else page

        if isinstance(self.output.channel, GuildChannel):
            return await self.output.edit(embed=safe_embed)

        else:
            await self.output.delete()

        self.output = await self.destination.send(embed=safe_embed)
        return self.output

    @staticmethod
    async def flush():
        """Helper method that will clear the user sessions list. Only call this if you know what you are doing."""
        sessions.clear()

    # Internal Methods
    async def _open(self):
        if not self.pages:
            return

        self._start_session()

        self.output = await self.destination.send(embed=self.page.as_safe_embed())
        self.input = self.ctx.message
        self.history.append(self.page.index)

        await self._cleanup_input()

    async def _post_next(self):
        """Sends a message after the `next` method is called. Closes the session if there is no callback on the next page."""
        if self.__class__.__name__ != 'PaginatedMenu':
            if self.page.on_next_event is None:
                await self.close_session()
                self.active = False

        self.history.append(self.page.index)
        await self.send_message(self.page)

    async def _execute_cancel(self):
        """Sends a cancellation message."""
        # we check if the page has a callback
        if self.page.on_cancel_event:
            return await self.page.on_cancel_event()

        await self._cleanup_output()
        await self.close_session()
        self.active = False

    async def close_session(self):
        """Remove the user from the active users list."""
        del sessions[(self.ctx.author.id, self.ctx.channel.id)]
        self.active = False

    async def _cleanup_input(self):
        """Deletes a Discord client user message."""
        if not self.command_message:
            if isinstance(self.input.channel, GuildChannel):
                await self.input.delete()

    async def _cleanup_output(self):
        """Deletes the Discord client bot message."""
        if not self.persist:
            self.output: Message
            await self.output.delete()
            self.output = None

    async def _execute_timeout(self):
        """Sends a timeout message."""
        # we check if the page has a callback
        if self.page.on_timeout_event:
            return await self.page.on_timeout_event()

        embed = Embed(title='Timed Out', description='You timed out at menu selection.')
        await self.send_message(embed)

        await self.close_session()
        self.active = False

    async def _get_input(self) -> Message:
        """Collects user input and places it into the input attribute."""
        try:
            message = await self.ctx.bot.wait_for('message', timeout=self.timeout, check=self._check_message)

        except asyncio.TimeoutError:
            if self.page.on_timeout_event:
                await self.page.on_timeout_event()

            else:
                await self._execute_timeout()

        else:
            return message

    def _check_message(self, m: Message) -> bool:
        """Returns true if the author is the person who responded and the channel is the same."""
        return (m.author == self.ctx.author
                and self.output.channel == m.channel)

    @staticmethod
    def _validate_pages(pages):
        """Checks that the Menu contains at least one pages."""
        if len(pages) == 0:
            raise PagesError(f'There must be at least one page in a menu. Expected at least 1, found {len(pages)}.')

    def _start_session(self):
        """Starts a new user session in the sessions storage. Raises a SessionError if the key already exists."""
        if (self.ctx.author.id, self.ctx.channel.id) in sessions.keys():
            raise SessionError(f'Duplicate session in channel [{self.ctx.channel.id}] for user [{self.ctx.author.id}].')
        else:
            sessions.update({(self.ctx.author.id, self.ctx.channel.id): self})
            return True



from dpymenus.base_menu import BaseMenu
from dpymenus.page import Page
from dpymenus.text_menu import TextMenu
from dpymenus.button_menu import ButtonMenu
from dpymenus.paginated_menu import PaginatedMenu
from dpymenus.poll import Poll



class MenuException(Exception):
    def __init__(self, message: str = None):
        self.message = message


class PagesError(MenuException):
    pass


class ButtonsError(MenuException):
    pass


class EventError(MenuException):
    pass


class SessionError(MenuException):
    pass



import logging
from typing import Dict, List, Union

from discord.abc import GuildChannel
from discord.ext.commands import Context

from dpymenus import BaseMenu
from dpymenus.constants import QUIT
from dpymenus.exceptions import SessionError


class TextMenu(BaseMenu):
    """
    Represents a text-based response menu.

    :param ctx: A reference to the command context.
    """

    def __init__(self, ctx: Context):
        super().__init__(ctx)

    def __repr__(self):
        return f'TextMenu(pages={[p.__str__() for p in self.pages]}, page={self.page}, timeout={self.timeout}, data={self.data})'

    @property
    def delay(self) -> float:
        return getattr(self, '_delay', 0.250)

    def set_delay(self, delay: float) -> 'TextMenu':
        """Sets the delay on when a users message will be deleted in guild channels. Returns itself for fluent-style chaining."""
        self._delay = delay

        return self

    @property
    def data(self) -> Dict:
        return getattr(self, '_data', {})

    def set_data(self, data: Dict) -> 'TextMenu':
        """Sets a dictionary up for persistent state data. Returns itself for fluent-style chaining."""
        self._data = data

        return self

    @property
    def normalized(self) -> bool:
        return getattr(self, '_normalized', False)

    def normalize_responses(self) -> 'TextMenu':
        """Strips all input data and ignores case when comparing strings with `response_is`. Returns itself for fluent-style chaining."""
        self._normalized = True

        return self

    def response_is(self, valid_response: Union[str, List[str]]) -> bool:
        """Helper method which checks if a users response is in the str or list of strings passed in."""
        response = self.input.content
        if self.normalized:
            response = ' '.join(response.lower().split())

        return any(response == res for res in valid_response) if isinstance(valid_response, List) else valid_response == response

    async def open(self):
        """The entry point to a new TextMenu instance; starts the main menu loop.
        Manages gathering user input, basic validation, sending messages, and cancellation requests."""
        try:
            await super()._open()
        except SessionError as exc:
            logging.info(exc.message)
        else:
            first_iter = True
            while self.active:
                if not first_iter and self.page.on_fail_event:
                    return await self.page.on_fail_event()

                first_iter = False

                self.input = await self._get_input()

                if self.input:
                    await self._cleanup_input()

                    if self._is_cancelled():
                        return await self._execute_cancel()

                    await self.page.on_next_event(self)

    # Internal Methods
    async def _cleanup_input(self):
        """Deletes a Discord client user message."""
        if isinstance(self.output.channel, GuildChannel):
            await self.input.delete(delay=self.delay)

    def _is_cancelled(self) -> bool:
        """Checks input for a cancellation string. If there is a match, it calls the ``menu.cancel()`` method and returns True."""
        if self.response_is(QUIT):
            return True
        return False



import setuptools

with open('README.md', 'r') as file:
    long_description = file.read()

setuptools.setup(
        name='dpymenus',
        version='1.3.0',
        author='Rob Wagner',
        author_email='rob.wagner@outlook.com',
        license='MIT',
        description='Simplified menus for discord.py developers.',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/robertwayne/dpymenus',
        packages=setuptools.find_packages(),
        classifiers=[
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Typing :: Typed',
            'Topic :: Communications :: Chat',
            'Intended Audience :: Developers',
            'Development Status :: 5 - Production/Stable',
        ],
        python_requires='>=3.7',
        install_requires=[
            'discord.py>=1.6.0',
            'emoji==0.6.0'
        ]
)
